/*
 * Copyright 2016 - 2023 Gary Talent (gary@drinkingtea.net). All rights reserved.
 */

#include <ctime>

#include <ox/logconn/logconn.hpp>
#include <ox/std/trace.hpp>
#include <ox/std/uuid.hpp>

#include <nostalgia/appmodules/appmodules.hpp>
#include <nostalgia/core/core.hpp>
#include <keel/media.hpp>

#include "lib/context.hpp"
#include "studioapp.hpp"

namespace nostalgia {

class StudioUIDrawer: public core::Drawer {
	private:
		StudioUI *m_ui = nullptr;
	public:
		explicit StudioUIDrawer(StudioUI *ui) noexcept: m_ui(ui) {
		}
	protected:
		void draw(core::Context*) noexcept final {
			m_ui->draw();
		}
};

static int updateHandler(core::Context *ctx) noexcept {
	auto sctx = core::applicationData<studio::StudioContext>(ctx);
	auto ui = dynamic_cast<StudioUI*>(sctx->ui);
	ui->update();
	return 16;
}

static void keyEventHandler(core::Context *ctx, core::Key key, bool down) noexcept {
	auto sctx = core::applicationData<studio::StudioContext>(ctx);
	auto ui = dynamic_cast<StudioUI*>(sctx->ui);
	ui->handleKeyEvent(key, down);
}

static ox::Error run(ox::UniquePtr<ox::FileSystem> fs) noexcept {
	oxRequireM(ctx, core::init(std::move(fs), "NostalgiaStudio"));
	core::setWindowTitle(ctx.get(), "Nostalgia Studio");
	core::setUpdateHandler(ctx.get(), updateHandler);
	core::setKeyEventHandler(ctx.get(), keyEventHandler);
	core::setConstantRefresh(ctx.get(), false);
	core::gl::setMainViewEnabled(false);
	studio::StudioContext studioCtx;
	core::setApplicationData(ctx.get(), &studioCtx);
	StudioUI ui(ctx.get());
	studioCtx.ui = &ui;
	StudioUIDrawer drawer(&ui);
	core::addCustomDrawer(ctx.get(), &drawer);
	auto err = core::run(ctx.get());
	core::removeCustomDrawer(ctx.get(), &drawer);
	return err;
}

static ox::Error run(int, const char**) noexcept {
	ox::trace::init();
	const auto time = std::time(nullptr);
	ox::UUID::seedGenerator({
		static_cast<uint64_t>(time),
		static_cast<uint64_t>(time << 1)
	});
	loadModules();
	return run(ox::UniquePtr<ox::FileSystem>(nullptr));
}

}

int main(int argc, const char **argv) {
#ifdef DEBUG
	ox::LoggerConn loggerConn;
	const auto loggerErr = loggerConn.initConn("Nostalgia Studio");
	if (loggerErr) {
		oxErrf("Could not connect to logger: {}\n", toStr(loggerErr));
	} else {
		ox::trace::setLogger(&loggerConn);
	}
#endif
	const auto err = nostalgia::run(argc, argv);
	oxAssert(err, "Something went wrong...");
	return static_cast<int>(err);
}
